//
// Created by martin on 17.03.24.
//

#ifndef ANIMTESTBED_ANIMGRAPHRESOURCE_H
#define ANIMTESTBED_ANIMGRAPHRESOURCE_H

#include "3rdparty/json/json.hpp"
#include "AnimGraphNodes.h"

struct AnimGraphBlendTree;
struct AnimGraphStateMachine;

struct AnimNodeResource {
  virtual ~AnimNodeResource() = default;

  std::string m_name;
  std::string m_node_type_name;
  AnimNode* m_anim_node = nullptr;
  NodeDescriptorBase* m_socket_accessor = nullptr;
  float m_position[2] = {0.f, 0.f};
};

static inline AnimNodeResource* AnimNodeResourceFactory(
    const std::string& node_type_name);

struct BlendTreeConnectionResource {
  size_t source_node_index = -1;
  std::string source_socket_name;
  size_t target_node_index = -1;
  std::string target_socket_name;
};

struct BlendTreeResource {
  std::vector<std::vector<size_t> > m_node_input_connection_indices;
  std::vector<std::vector<size_t> > m_node_output_connection_indices;

  ~BlendTreeResource() { CleanupNodes(); }

  void Reset() {
    CleanupNodes();

    m_connections.clear();

    m_node_input_connection_indices.clear();
    m_node_output_connection_indices.clear();
  }

  void CleanupNodes() {
    for (AnimNodeResource* node_resource : m_nodes) {
      delete node_resource->m_anim_node;
      delete node_resource->m_socket_accessor;
      delete node_resource;
    }

    m_nodes.clear();
  }

  void InitGraphConnectors() {
    AddNode(AnimNodeResourceFactory("BlendTreeSockets"));
    AnimNodeResource* output_node = GetGraphOutputNode();
    output_node->m_name = "Outputs";

    AddNode(AnimNodeResourceFactory("BlendTreeSockets"));
    AnimNodeResource* input_node = GetGraphInputNode();
    output_node->m_name = "Inputs";
  }

  [[nodiscard]] AnimNodeResource* GetGraphOutputNode() const {
    return m_nodes[0];
  }
  [[nodiscard]] AnimNodeResource* GetGraphInputNode() const {
    return m_nodes[1];
  }

  size_t GetNodeIndex(const AnimNodeResource* node_resource) const {
    for (size_t i = 0, n = m_nodes.size(); i < n; i++) {
      if (m_nodes[i] == node_resource) {
        return i;
      }
    }

    return -1;
  }

  [[maybe_unused]] size_t AddNode(AnimNodeResource* node_resource) {
    m_nodes.push_back(node_resource);
    m_node_input_connection_indices.emplace_back();
    m_node_output_connection_indices.emplace_back();
    return m_nodes.size() - 1;
  }

  [[nodiscard]] size_t GetNumNodes() const { return m_nodes.size(); }
  [[nodiscard]] AnimNodeResource* GetNode(size_t i) { return m_nodes[i]; }
  [[nodiscard]] const AnimNodeResource* GetNode(size_t i) const {
    return m_nodes[i];
  }
  [[nodiscard]] const std::vector<AnimNodeResource*>& GetNodes() const {
    return m_nodes;
  }

  [[nodiscard]] size_t GetNumConnections() const {
    return m_connections.size();
  }

  [[nodiscard]] BlendTreeConnectionResource* GetConnection(size_t i) {
    return &m_connections[i];
  }

  [[nodiscard]] const BlendTreeConnectionResource* GetConnection(
      size_t i) const {
    return &m_connections[i];
  }

  [[nodiscard]] const std::vector<BlendTreeConnectionResource>& GetConnections()
      const {
    return m_connections;
  }

  bool ConnectSockets(
      const AnimNodeResource* source_node,
      const std::string& source_socket_name,
      const AnimNodeResource* target_node,
      const std::string& target_socket_name);

  std::vector<Socket*> GetConstantNodeInputs(
      std::vector<NodeDescriptorBase*>& instance_node_descriptors) const {
    std::vector<Socket*> result;

    for (size_t i = 0; i < m_nodes.size(); i++) {
      for (size_t j = 0,
                  num_inputs = instance_node_descriptors[i]->m_inputs.size();
           j < num_inputs;
           j++) {
        Socket& input = instance_node_descriptors[i]->m_inputs[j];

        if (*input.m_reference.ptr_ptr == nullptr) {
          memcpy(
              &input.m_value,
              &m_nodes[i]->m_socket_accessor->m_inputs[j].m_value,
              sizeof(Socket::SocketValue));
          result.push_back(&input);
        }
      }
    }

    return result;
  }

  bool IsSocketConnected(
      const AnimNodeResource* source_node,
      const std::string& socket_name) {
    assert(false && "Not yet implemented");
    return false;
  }

  size_t GetNodeIndexForOutputSocket(const std::string& socket_name) const {
    for (size_t i = 0; i < m_connections.size(); i++) {
      const BlendTreeConnectionResource& connection = m_connections[i];
      if (connection.target_node_index == 0
          && connection.target_socket_name == socket_name) {
        return connection.source_node_index;
      }
    }

    std::cerr << "Error: could not find a node connected to output '"
              << socket_name << "'." << std::endl;

    return -1;
  }

  size_t GetNodeIndexForInputSocket(const std::string& socket_name) const {
    for (size_t i = 0; i < m_connections.size(); i++) {
      const BlendTreeConnectionResource& connection = m_connections[i];
      if (connection.source_node_index == 1
          && connection.source_socket_name == socket_name) {
        return connection.target_node_index;
      }
    }

    std::cerr << "Error: could not find a node connected to input '"
              << socket_name << "'." << std::endl;

    return -1;
  }

 private:
  std::vector<AnimNodeResource*> m_nodes;
  std::vector<BlendTreeConnectionResource> m_connections;
};

struct StateMachineTransitionResources {
  size_t source_state_index = -1;
  size_t target_state_index = -1;
  float blend_time = 0.f;
  bool sync_blend = false;
};

struct StateMachineResource {
  std::vector<AnimNodeResource> m_states;
  std::vector<StateMachineTransitionResources> m_transitions;
};

struct AnimGraphResource : AnimNodeResource {
  std::string m_graph_type_name;

  BlendTreeResource m_blend_tree_resource;
  typedef std::pair<const AnimNodeResource*, std::string> NodeSocketPair;
  typedef std::map<NodeSocketPair, int> NodeSocketDataOffsetMap;

  StateMachineResource m_state_machine_resource;

  void Clear() { m_blend_tree_resource.Reset(); }
  bool SaveToFile(const char* filename) const;
  bool LoadFromFile(const char* filename);

  void CreateBlendTreeInstance(AnimGraphBlendTree& result) const;
  void CreateStateMachineInstance(AnimGraphStateMachine& result) const;

 private:
  // BlendTree
  bool SaveBlendTreeResourceToFile(const char* filename) const;
  void CreateBlendTreeRuntimeNodeInstances(AnimGraphBlendTree& result) const;
  void PrepareBlendTreeIOData(
      AnimGraphBlendTree& instance,
      NodeSocketDataOffsetMap& node_offset_map) const;
  void CreateBlendTreeConnectionInstances(
      AnimGraphBlendTree& instance,
      NodeSocketDataOffsetMap& node_offset_map) const;
  void SetRuntimeNodeProperties(AnimGraphBlendTree& result) const;

  bool SaveStateMachineResourceToFile(const char* filename) const;
  bool LoadStateMachineResourceFromJson(nlohmann::json const& json_data);
};

static inline AnimNodeResource* AnimNodeResourceFactory(
    const std::string& node_type_name) {
  AnimNodeResource* result;

  if (node_type_name == "BlendTree") {
    AnimGraphResource* blend_tree_resource = new AnimGraphResource();
    blend_tree_resource->m_blend_tree_resource.InitGraphConnectors();
    result = blend_tree_resource;
  } else {
    result = new AnimNodeResource();
  }

  result->m_node_type_name = node_type_name;

  if (node_type_name == "BlendTreeSockets") {
    result->m_anim_node = AnimNodeFactory("BlendTree");
    result->m_socket_accessor = new NodeDescriptorBase();
  } else {
    result->m_anim_node = AnimNodeFactory(node_type_name);
    result->m_socket_accessor =
        AnimNodeDescriptorFactory(node_type_name, result->m_anim_node);
  }

  return result;
}

#endif  //ANIMTESTBED_ANIMGRAPHRESOURCE_H
